//+------------------------------------------------------------------+
//|                                                   Optimystic.mq4 |
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: On The Fly Self-Optimization of Expert Advisers |
//|                        https://www.mql5.com/en/blogs/post/683358 |
//+------------------------------------------------------------------+

#property library
#property strict

#define OPTIMYSTIC_LIBRARY

#include <Optimystic.mqh>
#include <RubbArray.mqh>

class Order
{
  private:
    static int tickets;
    
  public:
    
    int type;
    string symbol;
    double openPrice;
    datetime openTime;
    double lots;
    double takeProfit;
    double stopLoss;
    string comment;
    datetime expiration;
    int magicNumber;
    color arrow;

    int ticket;

    double closePrice;
    datetime closeTime;

    
    Order(datetime now, string s, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string cmnt = NULL, int magic = 0, datetime exp = 0, color clr = clrNONE)
    {
      symbol = s;
      type = cmd;
      lots = volume;
      openPrice = price;
      openTime = now;
      stopLoss = stoploss;
      takeProfit = takeprofit;
      comment = cmnt;
      magicNumber = magic;
      expiration = exp;
      arrow = clr;
      
      // deviation is ignored
      
      ticket = ++tickets;
    }

    double profit(double price, const int spread)
    {
      if(type == OP_BUY || type == OP_SELL)
      {
        if(closePrice != 0) price = closePrice;
        
        // MODE_TICKVALUE uses current exchange rates!!! should be rewritten using historic quotes,
        double ptv = MarketInfo(symbol, MODE_POINT) * MarketInfo(symbol, MODE_TICKVALUE) / MarketInfo(symbol, MODE_TICKSIZE);
        double pts = (type == OP_BUY ? +1 : -1) * (price - openPrice) / MarketInfo(symbol, MODE_POINT);
        
        return (pts - spread) * ptv * lots;
      }
      return 0;
    }
    
    double swap(datetime now) // unsupported
    {
      return 0;
    }
    
    //double commission() // unavailable here, moved to trade context
    //{
    //  return 0;
    //}
    
    
    void print(double price, int spread, datetime now)
    {
      int d = (int)MarketInfo(symbol, MODE_DIGITS);
      Print(ticket, "; ", (string)openTime, "; ", type, "; ", DoubleToString(lots, 2), "; ", symbol, " ", DoubleToString(openPrice, d), "; ", DoubleToString(stopLoss, d), "; ", DoubleToString(takeProfit, d), "; ", (string)closeTime, "; ", DoubleToString(closePrice, d), "; n/a; ", DoubleToString(swap(now), 2), "; ", DoubleToString(profit(price, spread), 2), "; ", comment, "; ", magicNumber, "; ", (string)expiration);
    }
    
    bool isMarketOrder() const
    {
      return type == OP_BUY || type == OP_SELL;
    }
    
};

int Order::tickets = 0;

class BarState
{
  public:
    datetime timestamp;
    double balance;
    double equity;
    
    BarState(const datetime dt, double b): timestamp(dt), balance(b), equity(b) {}
    BarState(const datetime dt, double b, double e): timestamp(dt), balance(b), equity(e) {}
};

enum PRICE_EMULATION_MODE {MODE_OPEN_PRICES, MODE_CONTROL_POINTS/*not implemented*/};

struct Statistics
{
  double deposit;
  double balance;
  
  double profit; // floating
  double plusProfit;
  double minusProfit;
  int plusCount;
  int minusCount;
  
  double drawdownAmount;
  double drawdownPercent;
  double sharpe;
  
  double getPerformance(const PERFORMANCE_ESTIMATOR type = ESTIMATOR_DEFAULT)
  {
    if(type == ESTIMATOR_DRAWDOWN_PERCENT)
    {
      return 100 - drawdownPercent; // inverted for maximization
    }
    else if(type == ESTIMATOR_PROFIT_FACTOR)
    {
      return minusProfit != 0 ? plusProfit / minusProfit : DBL_MAX;
    }
    else if(type == ESTIMATOR_WIN_PERCENT)
    {
      return plusCount + minusCount != 0 ? 100.0 * plusCount / (plusCount + minusCount) : 0;
    }
    else if(type == ESTIMATOR_AVERAGE_PROFIT)
    {
      return plusCount + minusCount != 0 ? AccountBalance() / (plusCount + minusCount) : 0;
    }
    else if(type == ESTIMATOR_TRADES_COUNT)
    {
      return plusCount + minusCount;
    }
    else if(type == ESTIMATOR_SHARPE)
    {
      return sharpe;
    }
    
    // default: (type == ESTIMATOR_DEFAULT || type == ESTIMATOR_NET_PROFIT)
    return balance - deposit;
  }
  
};

class OptimysticTradeContextImplementation: public OptimysticTradeContext
{
  private:
    int bar;
    double deposit;
    int spread;
    double commission;
    
    RubbArray<Order> orders;
    RubbArray<Order> history;
    Order *selected;

    RubbArray<BarState> income;
    Statistics stats;
    
    const PRICE_EMULATION_MODE priceMode;
    
    Order * findByTicket(int &ticket, RubbArray<Order> &array)
    {
      int n = array.size();
      for(int i = 0; i < n; i++)
      {
        Order *order = array[i];
        if(order.ticket == ticket)
        {
          ticket = i;
          return order;
        }
      }
      return NULL;
    }
    
    void calcDD()
    {
      double AbsDrawdown = 0, MaxProfit = 0, MaxDrawdown = 0, MaxPeak = 0, RelDrawdown = 0, Drawdown = 0;
      
      double total = income[0].equity;
      double initdepo = total;

      MaxProfit = total;
      
      int length = income.size();
      
      for(int i = 1; i < length; i++)
      {
        total = income[i].equity;
        
        if(AbsDrawdown < initdepo - total) AbsDrawdown = initdepo - total;
        if(total > MaxProfit) MaxProfit = total;
        if(MaxDrawdown < MaxProfit - total)
        {
          MaxDrawdown = MaxProfit - total;
          MaxPeak = MaxProfit;
          if(MaxPeak > 0)
          {
            double relative = 100 * MaxDrawdown / MaxPeak;
            if(RelDrawdown < relative)
            {
              RelDrawdown = relative;
              Drawdown = MaxDrawdown;
            }
          }
        }
      }
      
      stats.drawdownAmount = Drawdown;
      stats.drawdownPercent = RelDrawdown;
    }
    
    void calcSharpeRatio(const bool direct = true)
    {
      stats.sharpe = 0;
    
      double AHPR = 0;
      double Std = 0;
      
      int i, limit = income.size();
      if(limit < 3) return;

      double HPR[];
      ArrayResize(HPR, limit);

      int N = limit - 1;
      if(direct)
      {
        for(i = 1; i < limit; i++)
        {
          if(income[i - 1].equity != 0) 
          {
            HPR[i - 1] = (income[i].equity - income[i - 1].equity) / income[i - 1].equity;
            AHPR += HPR[i - 1];
          }
        }
      }
      else
      {
        for(i = limit - 2; i>=0; i--)
        {
          if(income[i + 1].equity != 0) 
          {
            HPR[i] = (income[i].equity - income[i + 1].equity) / income[i + 1].equity;
            AHPR += HPR[i];
          }
        }
      }
      AHPR = AHPR / N;
     
      for(i = 0; i < N - 1; i++)
      {
        Std += (AHPR - HPR[i]) * (AHPR - HPR[i]);
      }
      Std = MathSqrt(Std / (N - 1));
      if(Std == 0) return;
      stats.sharpe = AHPR / Std;
    }
    
    
  public:
  
    Statistics getStatistics() const
    {
      return stats;
    }
    
    OptimysticTradeContextImplementation(): priceMode(MODE_OPEN_PRICES)
    {
    }
    
    void setBar(const int b)
    {
      bar = b;
      income << new BarState(TimeCurrent(), income.top().balance);
    }
    
    void initialize(const double d, const int s, const double c)
    {
      deposit = d;
      spread = s;
      commission = c;

      stats.profit = 0;
      stats.plusProfit = stats.minusProfit = 0;
      stats.plusCount = stats.minusCount = 0;
      stats.deposit = d;
      stats.balance = d;
      stats.drawdownAmount = 0;
      stats.drawdownPercent = 0;
      stats.sharpe = 0;
      
      income.clear();
      income << new BarState(TimeCurrent(), deposit);
    }

    void activatePendingOrdersAndStops()
    {
      RubbArray<Order> toClose;
      RubbArray<Order> toDelete;
      
      int n = orders.size();
      for(int i = 0; i < n; i++)
      {
        Order *order = orders[i];
        double spreadValue = spread * MarketInfo(order.symbol, MODE_POINT);
        double low = ::iLow(order.symbol, Period(), bar);
        double high = ::iHigh(order.symbol, Period(), bar);
        
        if(order.type == OP_SELLSTOP)
        {
          if(low <= order.openPrice)
          {
            order.type = OP_SELL;
            order.openTime = TimeCurrent();
          }
        }
        else if(order.type == OP_SELLLIMIT)
        {
          if(high >= order.openPrice)
          {
            order.type = OP_SELL;
            order.openTime = TimeCurrent();
          }
        }
        else if(order.type == OP_BUYLIMIT)
        {
          if(low + spreadValue <= order.openPrice)
          {
            order.type = OP_BUY;
            order.openTime = TimeCurrent();
          }
        }
        else if(order.type == OP_BUYSTOP)
        {
          if(high + spreadValue >= order.openPrice)
          {
            order.type = OP_BUY;
            order.openTime = TimeCurrent();
          }
        }
        else if(order.type == OP_BUY)
        {
          if(low <= order.stopLoss)
          {
            order.closePrice = order.stopLoss;
            order.comment += "[sl]";
            toClose << order;
          }
          else if(order.takeProfit != 0 && high >= order.takeProfit)
          {
            order.closePrice = order.takeProfit;
            order.comment += "[tp]";
            toClose << order;
          }
        }
        else if(order.type == OP_SELL)
        {
          if(order.stopLoss != 0 && high + spreadValue >= order.stopLoss)
          {
            order.closePrice = order.stopLoss;
            order.comment += "[sl]";
            toClose << order;
          }
          else if(low + spreadValue <= order.takeProfit)
          {
            order.closePrice = order.takeProfit;
            order.comment += "[tp]";
            toClose << order;
          }
        }
        
        if(order.type > OP_SELL)
        {
          if(order.expiration > 0 && TimeCurrent() >= order.expiration)
          {
            order.closePrice = iOpen(order.symbol, Period(), 0) + (order.type == OP_BUY ? spreadValue : 0);
            order.closeTime = TimeCurrent();
            order.comment += "[exp]";
            toDelete << order;
          }
        }
      }
      
      while(toClose.size() > 0)
      {
        Order *order = toClose >> 0;
        if(order != NULL)
        {
          OrderClose(order.ticket, order.lots, order.closePrice, 0);
        }
      }
      
      while(toDelete.size() > 0)
      {
        Order *order = toDelete >> 0;
        if(order != NULL)
        {
          OrderDelete(order.ticket);
        }
      }
      
    }
    
    void closeAll()
    {
      int n = orders.size();
      for(int i = n - 1; i >= 0; i--)
      {
        Order *order = orders[i];
        OrderClose(order.ticket, order.lots, iOpen(order.symbol, Period(), 0), 0);
      }
      income.top().equity = income.top().balance;
      stats.profit = 0;
    }
    
    void calculateEquity()
    {
      income.top().equity = income.top().balance;
      int n = orders.size();
      stats.profit = 0;
      for(int i = 0; i < n; i++)
      {
        Order *order = orders[i];
        // use high/low prices for realistic drawdown estimation
        if(order.type == OP_BUY)
        {
          stats.profit += order.profit(::iLow(order.symbol, Period(), bar), spread) + order.swap(TimeCurrent()) + order.lots * commission;
        }
        else if(order.type == OP_SELL)
        {
          stats.profit += order.profit(::iHigh(order.symbol, Period(), bar), spread) + order.swap(TimeCurrent()) + order.lots * commission;
        }
      }
      income.top().equity += stats.profit;
    }
    
    void calculateStats()
    {
      stats.balance = AccountBalance();
      calcDD();
      calcSharpeRatio();
    }
    
    double getPerformance(const PERFORMANCE_ESTIMATOR type = ESTIMATOR_DEFAULT)
    {
      return stats.getPerformance(type);
    }


    //    I N T E R F A C E
    
    virtual double AccountBalance()
    {
      return income.size() > 0 ? income.top().balance : 0;
    }
    
    virtual double AccountEquity()
    {
      return income.size() > 0 ? income.top().equity : 0;
    }
    
    virtual double AccountProfit()
    {
      return stats.profit;
    }
    
    virtual double AccountFreeMargin()
    {
      return AccountEquity() - AccountMargin();
    }
    
    virtual double AccountFreeMarginCheck(string symbol, int cmd, double volume)
    {
      double margin = MarketInfo(symbol, MODE_MARGINREQUIRED) * volume;
      return AccountFreeMargin() - margin;
    }
    
    virtual double AccountMargin()
    {
      double margin = 0;
      int n = orders.size();
      for(int i = 0; i < n; i++)
      {
        Order *order = orders[i];
        margin += MarketInfo(order.symbol, MODE_MARGINREQUIRED) * order.lots;
      }
      return margin;
    }

    
    
    virtual datetime TimeCurrent()
    {
      return iTime(::Symbol(), ::Period(), bar);
    }
    
    virtual int OrdersHistoryTotal()
    {
      return history.size();
    }
      
    virtual int OrdersTotal()
    {
      return orders.size();
    }
    
    virtual int OrderSend(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string comment = NULL, int magic = 0, datetime expiration = 0, color arrow = clrNONE)
    {
      Order *order = new Order(TimeCurrent(), symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magic, expiration, arrow);
      orders << order;
      return order.ticket;
    }
    
    virtual bool OrderClose(int ticket, double lots, double price, int slippage, color arrow = clrNONE)
    {
      Order *order = findByTicket(ticket, orders);
      if(order != NULL && order.isMarketOrder())
      {
        orders >> ticket;
        history << order;
        // TODO: if lots < order.lots do partial close:
        // create new order with neworder.lots = order.lots - lots, copy type, symbol, openPrice and openTime from old order
        // change old order.lots = lots
        order.closePrice = price;
        order.closeTime = TimeCurrent();
        
        double amount = order.profit(price, spread) + order.swap(order.closeTime) + order.lots * commission;
        income.top().balance += amount;
        
        if(amount > 0)
        {
          stats.plusProfit += amount;
          stats.plusCount++;
        }
        else
        {
          stats.minusProfit -= amount;
          stats.minusCount++;
        }

        return true;
      }
      return false;
    }
    
    virtual bool OrderDelete(int ticket, color arrow = clrNONE)
    {
      Order *order = findByTicket(ticket, orders);
      if(order != NULL && !order.isMarketOrder())
      {
        orders >> ticket;
        history << order;
        return true;
      }
      return false;
    }
    
    virtual bool OrderModify(int ticket, double price, double stoploss, double takeprofit, datetime expiration, color arrow = clrNONE)
    {
      Order *order = findByTicket(ticket, orders);
      if(order != NULL)
      {
        if(!order.isMarketOrder())
        {
          order.openPrice = price;
          order.expiration = expiration;
        }
        order.stopLoss = stoploss;
        order.takeProfit = takeprofit;
        order.arrow = arrow;

        return true;
      }
      return false;
    }
    
    virtual bool OrderSelect(int index, int select, int pool = MODE_TRADES)
    {
      if(select == SELECT_BY_POS)
      {
        selected = pool == MODE_TRADES ? orders[index] : history[index];
      }
      else // SELECT_BY_TICKET
      {
        if(pool == MODE_TRADES)
        {
          selected = findByTicket(index, orders);
        }
        else
        {
          selected = findByTicket(index, history);
        }
      }
      
      return selected != NULL;
    }
    
    virtual double OrderClosePrice()
    {
      if(selected != NULL) return selected.closePrice;
      return 0;
    }
    
    virtual datetime OrderCloseTime()
    {
      if(selected != NULL) return selected.closeTime;
      return 0;
    }
    
    virtual string OrderComment()
    {
      if(selected != NULL) return selected.comment;
      return "";
    }
    
    virtual datetime OrderExpiration()
    {
      if(selected != NULL) return selected.expiration;
      return 0;
    }
    
    virtual double OrderLots()
    {
      if(selected != NULL) return selected.lots;
      return 0;
    }
    
    virtual int OrderMagicNumber()
    {
      if(selected != NULL) return selected.magicNumber;
      return 0;
    }
    
    virtual double OrderOpenPrice()
    {
      if(selected != NULL) return selected.openPrice;
      return 0;
    }
    
    virtual datetime OrderOpenTime()
    {
      if(selected != NULL) return selected.openTime;
      return 0;
    }
    
    virtual void OrderPrint()
    {
      if(selected != NULL) selected.print(iOpen(selected.symbol, Period(), 0), spread, TimeCurrent());
    }
    
    virtual double OrderProfit()
    {
      if(selected != NULL) return selected.profit(iOpen(selected.symbol, Period(), 0), spread);
      return 0;
    }
    
    virtual double OrderStopLoss()
    {
      if(selected != NULL) return selected.stopLoss;
      return 0;
    }
    
    virtual double OrderSwap()
    {
      if(selected != NULL) return selected.swap(TimeCurrent());
      return 0;
    }
    
    virtual double OrderCommission()
    {
      if(selected != NULL) return selected.lots * commission;
      return 0;
    }
    
    virtual string OrderSymbol()
    {
      if(selected != NULL) return selected.symbol;
      return "";
    }
    
    virtual double OrderTakeProfit()
    {
      if(selected != NULL) return selected.takeProfit;
      return 0;
    }
    
    virtual int OrderTicket()
    {
      if(selected != NULL) return selected.ticket;
      return 0;
    }
    
    virtual int OrderType()
    {
      if(selected != NULL) return selected.type;
      return 0;
    }
    

    
    virtual double iOpen(string symbol, int tf, int offset)
    {
      int b = bar + offset;
      if(symbol != Symbol())
      {
        b = ::iBarShift(symbol, tf, Time[b]);
      }
      return ::iOpen(symbol, tf, b);
    }
    
    virtual double iHigh(string symbol, int tf, int offset)
    {
      int b = bar + offset;
      if(symbol != Symbol())
      {
        b = ::iBarShift(symbol, tf, Time[b]);
      }
      if(offset == 0 && priceMode == MODE_OPEN_PRICES) return ::iOpen(symbol, tf, b);
      return ::iHigh(symbol, tf, b);
    }
    
    virtual double iLow(string symbol, int tf, int offset)
    {
      int b = bar + offset;
      if(symbol != Symbol())
      {
        b = ::iBarShift(symbol, tf, Time[b]);
      }
      if(offset == 0 && priceMode == MODE_OPEN_PRICES) return ::iOpen(symbol, tf, b);
      return ::iLow(symbol, tf, b);
    }
    
    virtual double iClose(string symbol, int tf, int offset)
    {
      int b = bar + offset;
      if(symbol != Symbol())
      {
        b = ::iBarShift(symbol, tf, Time[b]);
      }
      if(offset == 0 && priceMode == MODE_OPEN_PRICES) return ::iOpen(symbol, tf, b);
      return ::iClose(symbol, tf, b);
    }
    
    virtual datetime iTime(string symbol, int tf, int offset)
    {
      int b = bar + offset;
      if(symbol != Symbol())
      {
        b = ::iBarShift(symbol, tf, Time[b]);
      }
      return ::iTime(symbol, tf, b);
    }
    
    virtual int iBarShift(string symbol, int tf, datetime dt, bool exact = false)
    {
      int b = ::iBarShift(symbol, tf, dt, exact);
      if(b != -1 && b >= bar)
      {
        b -= bar;
      }
      return b;
    }
};

class OptimysticParameterImplementation: public OptimysticParameter
{
  protected:
    string name;
    double min;
    double max;
    double step;
    double value;
    double best;
    bool enabled;
  
  public:
    OptimysticParameterImplementation(const string _name, const double _min, const double _max, const double _step)
    {
      name = _name;
      min = MathMin(_min, _max);
      max = MathMax(_max, _min);
      enabled = true;
      step = _step <= 0 ? 1 : _step;
      best = EMPTY_VALUE;
    }
    
    void setValue(const double v)
    {
      value = v;
    }
    
    void increment()
    {
      value += step;
    }
    
    void markAsBest()
    {
      best = value;
    }

    virtual string getName() const
    {
      return name;
    }
    virtual double getMin() const
    {
      return min;
    }
    virtual double getMax() const
    {
      return max;
    }
    virtual double getStep() const
    {
      return step;
    }
    virtual double getValue() const
    {
      return value;
    }
    virtual double getBest() const
    {
      return best;
    }
    virtual bool getEnabled() const
    {
      return enabled;
    }
    virtual void setEnabled(const bool e)
    {
      enabled = e;
    }
    
};

class OptimysticImplementation: public Optimystic
{
  private:
    OptimysticCallback *callback;
    OptimysticTradeContextImplementation *context;
    bool enabled;
    int rangeInBars;
    int resetInBars;
    int forwardBars;
    datetime lastOptimization;
    PERFORMANCE_ESTIMATOR estimator;
    double deposit;
    RubbArray<OptimysticParameterImplementation> parameters;
    int parameterSpace[];
    int parameterCursor[];
    double best;
    Statistics bestStats;
    bool isOptimizationOn;
    
    
    int spread; // TODO: multicurrency support
    double commission;
    
  protected:
    void resetParameters()
    {
      int n = parameters.size();
      ArrayResize(parameterSpace, n);
      ArrayInitialize(parameterSpace, 0);
      ArrayResize(parameterCursor, n);
      ArrayInitialize(parameterCursor, 0);
      for(int i = 0; i < n; i++)
      {
        parameterSpace[i] = (int)((parameters[i].getMax() - parameters[i].getMin()) / parameters[i].getStep());
        parameterSpace[i]++;
        if(parameters[i].getEnabled()) parameters[i].setValue(parameters[i].getMin());
      }
    }
  
    bool iterateParameters()
    {
      int n = parameters.size();
      for(int i = 0; i < n; i++)
      {
        if(!parameters[i].getEnabled()) continue;
        if(parameterCursor[i] < parameterSpace[i] - 1)
        {
          parameterCursor[i]++;
          parameters[i].increment();
          return true;
        }
        else
        {
          parameterCursor[i] = 0;
          parameters[i].setValue(parameters[i].getMin());
        }
      }
      return false;
    }
    
    void storeBestParameters()
    {
      int n = parameters.size();
      for(int i = 0; i < n; i++)
      {
        parameters[i].markAsBest();
      }
    }
  
    bool optimize()
    {
      if(!callback.onBeforeOptimization()) return false;
      Print("Starting on-the-fly optimization at ", TimeToString(Time[rangeInBars]));
      
      if(spread == 0) spread = (int)MarketInfo(Symbol(), MODE_SPREAD);
      best = 0;
      int counter = 0;
      resetParameters();
      do
      {
        if(!callback.onApplyParameters()) return false;
        isOptimizationOn = true;
        context.initialize(deposit, spread, commission);
        // build balance and equity curves on the fly
        for(int i = rangeInBars; i >= forwardBars; i--)
        {
          context.setBar(i);
          // open prices only (in future can generate control points for indicatorless EAs)
          callback.trade(i, iOpen(Symbol(), Period(), i) + spread * Point, iOpen(Symbol(), Period(), i));
          context.activatePendingOrdersAndStops();
          context.calculateEquity();
        }
        context.closeAll();

        // calculate estimators
        context.calculateStats();
        
        double gain;
        if(estimator == ESTIMATOR_CUSTOM)
        {
          gain = callback.onCustomEstimate();
          if(gain == EMPTY_VALUE)
          {
            Print("onCustomEstimate is EMPTY! Optimization aborted.");
            return false;
          }
        }
        else
        {
          gain = context.getPerformance(estimator);
        }
        
        // if estimator is better than previous, then save current parameters
        if(gain > best && context.getPerformance(ESTIMATOR_TRADES_COUNT) > 0)
        {
          best = gain;
          bestStats = context.getStatistics();
          storeBestParameters();
        }
        callback.onReadyParameters();
        isOptimizationOn = false;
        counter++;
      }
      while(iterateParameters() && !IsStopped());

      Print((string)counter + " optimization passes completed");
      // TODO: forward test from forwardBars-1 to 0 for best parameters
      
      callback.onAfterOptimization();
      
      lastOptimization = TimeCurrent();
      return true;
    }

  public:
    OptimysticImplementation(OptimysticCallback &cb)
    {
      callback = GetPointer(cb);
      context = new OptimysticTradeContextImplementation();
      estimator = ESTIMATOR_DEFAULT;
    }
    
    ~OptimysticImplementation()
    {
      delete context;
    }
    
    virtual void setEnabled(const bool _enabled)
    {
      enabled = _enabled;
    }

    virtual OptimysticParameter * addParameter(const string name, const double min, const double max, const double step)
    {
      parameters << new OptimysticParameterImplementation(name, min, max, step);
      return parameters.top();
    }

    virtual OptimysticParameter *getParameter(const int i) const
    {
      return parameters[i];
    }
    
    virtual void setHistoryDepth(const int size, const TIME_UNIT unit)
    {
      if(unit == UNIT_DAY)
      {
        // convert days to bars
        rangeInBars = size * (PeriodSeconds(PERIOD_D1) / PeriodSeconds());
      }
      else
      {
        rangeInBars = size;
      }
    }
    
    virtual void setReoptimizationPeriod(const int size, const TIME_UNIT unit)
    {
      if(unit == UNIT_DAY)
      {
        // convert days to bars
        resetInBars = size * (PeriodSeconds(PERIOD_D1) / PeriodSeconds());
      }
      else
      {
        resetInBars = size;
      }
      
      if(resetInBars < 1) resetInBars = 1;
    }
    
    virtual void setOptimizationCriterion(const PERFORMANCE_ESTIMATOR type)
    {
      estimator = type;
    }
    
    virtual void setDeposit(const double d)
    {
      deposit = d;
    }
    
    virtual void setSpread(const int points = 0)
    {
      spread = points;
    }
    
    virtual void setCommission(const double c)
    {
      commission = c;
    }
    
    virtual void onTick()
    {
      static datetime lastBar;
      if(enabled)
      {
        if(lastBar != Time[0] && TimeCurrent() > lastOptimization + PeriodSeconds() * resetInBars)
        {
          callback.setContext(context); // virtual trading
          optimize();
        }
      }
      callback.setContext(NULL); // real trading
      callback.trade(0, Ask, Bid);
      lastBar = Time[0];
    }
    
    virtual double getOptimalPerformance(PERFORMANCE_ESTIMATOR type = ESTIMATOR_DEFAULT)
    {
      if(type == ESTIMATOR_DEFAULT) type = estimator;
      if(type == ESTIMATOR_CUSTOM) return best;
      return isOptimizationOn ? context.getStatistics().getPerformance(type) : bestStats.getPerformance(type);
    }
    
};


Optimystic *createInstance(OptimysticCallback &callback) export
{
  return new OptimysticImplementation(callback);
}
